/*
** Module   : registry.c
** Abstract : implementation for Windows registry access functions.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130618 Created initial version to implement Windows specifics for registry related
**                  environment sccess operations.
*/

#include <windows.h>

#include "com_goldencode_p2j_util_Registry.h"

/**
* Helper function that will return the full name of the registry key for given section, base and
* key names. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section to scan.
* @param  key 
*         The name of the registry key within given section.
* @param  isWrite 
*         Flag to indicate whether we need to open key for writing or reading only.
* @param  pHKey
*         Pointer to the variable that will receive the handle value of the registry key.
* 
* @return A string representing the registry key name and the key handle value.
*/
char * getKeyNameAndHandle(JNIEnv* env, jstring base, jstring section, jstring key, BOOL isWrite,
                           HKEY * pHKey)
{
   char * name = NULL;
   char * keyPath = NULL;
   HKEY hBaseKey = 0;

   // Construct initial parameters for Windows API call
   const char *base_name = (*env)->GetStringUTFChars(env, base, NULL);
   const char *sect_name = (*env)->GetStringUTFChars(env, section, NULL);
   const char *key_name  = (*env)->GetStringUTFChars(env, key, NULL);
   
   // Reconstruct the handle to the registry hive
   hBaseKey = HKEY_CURRENT_USER;
   // TODO: Implement handling the other cases of base_name variable
   
   int keyLength = 0;
   // Check the memory size to hold the full key path name
   // TODO: Implement handling different missing parts either section or key name combinations
   if (sect_name != NULL && strlen(sect_name) > 0)
   {
      keyLength += strlen(sect_name);
   }
   // One more char for final null terminator string
   keyLength += 1;
   
   // Allocate memory to registry path name
   keyPath = malloc(sizeof(char) * keyLength);
   // And initialize with empty
   strcpy(keyPath, "");
   // Form the final key path
   if (sect_name != NULL && strlen(sect_name) > 0)
   {
      strcat(keyPath, sect_name);
   }   

   // Extract the key name itself
   keyLength = 0;
   if (key_name != NULL && strlen(key_name) > 0)
   {
      keyLength += strlen(key_name);
   }
   // One more char for final null terminator string
   keyLength += 1;
   
   // Allocate memory to registry key name
   name = malloc(sizeof(char) * keyLength);
   // And initialize with empty
   strcpy(name, "");
   if (key_name != NULL && strlen(key_name) > 0)
   {
      strcat(name, key_name);
   }

   printf("Key path and name to get handle: %s %s\n", keyPath, name);
   fflush(stdout);
   DWORD rc;
   if (isWrite) // Create or reopen subkey to write access
   {
      DWORD dwDisposition;
      rc = RegCreateKeyEx(hBaseKey, keyPath, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL,
                          pHKey, &dwDisposition);
   }
   else         // Open subkey to read only access
   {
      rc = RegOpenKeyEx(hBaseKey, keyPath, 0, KEY_READ, pHKey);
   }
   
   // Release Java memory
   (*env)->ReleaseStringUTFChars(env, key, key_name);
   (*env)->ReleaseStringUTFChars(env, section, sect_name);
   (*env)->ReleaseStringUTFChars(env, base, base_name);
   
   // Free local memory
   free(keyPath);
   // The registry key was created or opened
   if (rc == ERROR_SUCCESS)
   {
      printf("Key create-open success, handle: %p\n", *pHKey);
      fflush(stdout);
      // Remember to release allocated memory in caller in case of success
      return name;
   }
   else // Creation or opening failure, clean up resources   
   {
      printf("Key create-open failure, RC= %lu\n", rc);
      fflush(stdout);
      // No name value will be returned
      free(name);
      RegCloseKey(*pHKey);
      return NULL;
   }
}

/**
* Utility method that will return the value of the registry key for given section name. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  cls 
*         reference to the Java class who called this function
*         (from static context)
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section to scan.
* @param  key 
*         The name of the registry key within given section whose value to get.
* 
* @return A string representing the registry key value if exists or NULL pointer if the key does
*         not exist in the Windows registry.
*/
JNIEXPORT jstring JNICALL
   Java_com_goldencode_p2j_util_Registry_getKeyVal(JNIEnv* env,
                                                   jclass  cls,
                                                   jstring base,
                                                   jstring section,
                                                   jstring key)
{
   char * result = NULL;

   // Reconstruct the handle to the registry hive and full key name
   HKEY hKey = 0;
   // Get registry key for reading
   char * keyName = getKeyNameAndHandle(env, base, section, key, FALSE, &hKey);
   
   DWORD rc;
   DWORD cbData = 0;
   
   // Find out how much memory we need to store the result
   printf("Key name to read from: %p %s\n", hKey, keyName);
   fflush(stdout);
   rc = RegQueryValueEx(hKey, keyName, NULL, NULL, NULL, &cbData);
   if (rc == ERROR_SUCCESS && cbData > 0 && keyName != NULL)
   {
      result = malloc(sizeof(char) * cbData);
      if (result != NULL)
      {
         rc = RegQueryValueEx(hKey, keyName, NULL, NULL, (LPBYTE)result, &cbData);
      }
   }

   // Key name was valid
   if (keyName != NULL)
   {
      // Release local memory
      free(keyName);
      // And close opened registry key
      RegCloseKey(hKey);
   }

   if (result != NULL)
   {
      printf("Got key value: %s\n", result);
      fflush(stdout);
      // We need this additional variable to be able to free local memory allocated for
      // Windows API call and to prevent memory leak
      jstring res = (*env)->NewStringUTF(env, result);
      free(result);
      return res;
   }
   else
   {
      return NULL;
   }   
}

/**
* Utility method that will set the new value for the registry key for given section name. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  cls 
*         reference to the Java class who called this function
*         (from static context)
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section to scan.
* @param  key 
*         The name of the registry key within given section whose value to change.
* @param  value 
*         The new value for the registry entry to set for given section and given key.
* 
* @return TRUE in case of success, FALSE otherwise.
*/
JNIEXPORT jboolean JNICALL
   Java_com_goldencode_p2j_util_Registry_setKeyVal(JNIEnv* env,
                                                   jclass  cls,
                                                   jstring base,
                                                   jstring section,
                                                   jstring key,
                                                   jstring value)
{
   jboolean res = FALSE;
   // Construct initial parameters for Windows API call
   const char *key_value = (*env)->GetStringUTFChars(env, value, NULL);

   // Reconstruct the handle to the registry hive and full key name
   HKEY hKey = 0;
   // Get registry key for writing
   char * keyName = getKeyNameAndHandle(env, base, section, key, TRUE, &hKey);
   
   DWORD rc;
   DWORD cbData = 0;
   
   printf("Key name to modify: %p %s\n", hKey, keyName);
   fflush(stdout);
   if (key_value != NULL && strlen(key_value) > 0 && keyName != NULL)
   {
      // The size of data to be stored should include null terminator for string data
      cbData = strlen(key_value) + 1;
      rc = RegSetValueEx(hKey, keyName, 0, REG_SZ, (LPBYTE)key_value, cbData);
      if (rc == ERROR_SUCCESS)
      {
         printf("Set key value: %s\n", key_value);
         fflush(stdout);
         res = TRUE;
      }
   }
   
   // Key name was valid
   if (keyName != NULL)
   {
      // Release local memory
      free(keyName);
      // And close opened registry key
      RegCloseKey(hKey);
   }
   // Release Java memory
   (*env)->ReleaseStringUTFChars(env, value, key_value);
   
   return res;
}

